#version 110


#extension GL_ARB_shading_language_include : require
#extension GL_ARB_uniform_buffer_object : require


#include "all_green.h"


uniform float blue;
uniform vec4 color;
layout(std140) uniform small
{
    vec4 value1;
};


void main()
{
    gl_FragColor = vec4(red, green, blue, color.a);
}
